@@ -1,50 +1,51 @@
 """ core implementation of testing process: init, session, runtest loop. """
-import enum
+import argparse
 import fnmatch
 import functools
 import importlib
 import os
 import sys
+from typing import Callable
 from typing import Dict
+from typing import FrozenSet
+from typing import Iterator
+from typing import List
+from typing import Optional
+from typing import Sequence
+from typing import Set
+from typing import Tuple
+from typing import Union
 
 import attr
 import py
 
 import _pytest._code
 from _pytest import nodes
+from _pytest.compat import overload
+from _pytest.compat import TYPE_CHECKING
+from _pytest.config import Config
 from _pytest.config import directory_arg
+from _pytest.config import ExitCode
 from _pytest.config import hookimpl
 from _pytest.config import UsageError
+from _pytest.config.argparsing import Parser
 from _pytest.fixtures import FixtureManager
 from _pytest.outcomes import exit
+from _pytest.pathlib import Path
+from _pytest.reports import CollectReport
+from _pytest.reports import TestReport
 from _pytest.runner import collect_one_node
 from _pytest.runner import SetupState
 
 
-class ExitCode(enum.IntEnum):
-    """
-    .. versionadded:: 5.0
+if TYPE_CHECKING:
+    from typing import Type
+    from typing_extensions import Literal
 
-    Encodes the valid exit codes by pytest.
-
-    Currently users and plugins may supply other exit codes as well.
-    """
-
-    #: tests passed
-    OK = 0
-    #: tests failed
-    TESTS_FAILED = 1
-    #: pytest was interrupted
-    INTERRUPTED = 2
-    #: an internal error got in the way
-    INTERNAL_ERROR = 3
-    #: pytest was misused
-    USAGE_ERROR = 4
-    #: pytest couldn't find tests
-    NO_TESTS_COLLECTED = 5
+    from _pytest.python import Package
 
 
-def pytest_addoption(parser):
+def pytest_addoption(parser: Parser) -> None:
     parser.addini(
         "norecursedirs",
         "directory patterns to avoid for recursion",
@@ -66,7 +67,7 @@ def pytest_addoption(parser):
         dest="maxfail",
         const=1,
         help="exit instantly on first error or failed test.",
-    ),
+    )
     group._addoption(
         "--maxfail",
         metavar="num",
@@ -76,6 +77,11 @@ def pytest_addoption(parser):
         default=0,
         help="exit after first num failures or errors.",
     )
+    group._addoption(
+        "--strict-config",
+        action="store_true",
+        help="any warnings encountered while parsing the `pytest` section of the configuration file raise errors.",
+    )
     group._addoption(
         "--strict-markers",
         "--strict",
@@ -113,7 +119,7 @@ def pytest_addoption(parser):
         "--co",
         action="store_true",
         help="only collect tests, don't execute them.",
-    ),
+    )
     group.addoption(
         "--pyargs",
         action="store_true",
@@ -135,10 +141,8 @@ def pytest_addoption(parser):
         "--deselect",
         action="append",
         metavar="nodeid_prefix",
-        help="deselect item during collection (multi-allowed).",
+        help="deselect item (via node id prefix) during collection (multi-allowed).",
     )
-    # when changing this to --conf-cut-dir, config.py Conftest.setinitial
-    # needs upgrading as well
     group.addoption(
         "--confcutdir",
         dest="confcutdir",
@@ -169,12 +173,21 @@ def pytest_addoption(parser):
         default=False,
         help="Don't ignore tests in a local virtualenv directory",
     )
+    group.addoption(
+        "--import-mode",
+        default="prepend",
+        choices=["prepend", "append", "importlib"],
+        dest="importmode",
+        help="prepend/append to sys.path when importing test modules and conftest files, "
+        "default is to prepend.",
+    )
 
     group = parser.getgroup("debugconfig", "test session debugging and configuration")
     group.addoption(
         "--basetemp",
         dest="basetemp",
         default=None,
+        type=validate_basetemp,
         metavar="dir",
         help=(
             "base temporary directory for this test run."
@@ -183,9 +196,39 @@ def pytest_addoption(parser):
     )
 
 
-def wrap_session(config, doit):
+def validate_basetemp(path: str) -> str:
+    # GH 7119
+    msg = "basetemp must not be empty, the current working directory or any parent directory of it"
+
+    # empty path
+    if not path:
+        raise argparse.ArgumentTypeError(msg)
+
+    def is_ancestor(base: Path, query: Path) -> bool:
+        """ return True if query is an ancestor of base, else False."""
+        if base == query:
+            return True
+        for parent in base.parents:
+            if parent == query:
+                return True
+        return False
+
+    # check if path is an ancestor of cwd
+    if is_ancestor(Path.cwd(), Path(path).absolute()):
+        raise argparse.ArgumentTypeError(msg)
+
+    # check symlinks for ancestors
+    if is_ancestor(Path.cwd().resolve(), Path(path).resolve()):
+        raise argparse.ArgumentTypeError(msg)
+
+    return path
+
+
+def wrap_session(
+    config: Config, doit: Callable[[Config, "Session"], Optional[Union[int, ExitCode]]]
+) -> Union[int, ExitCode]:
     """Skeleton command line program"""
-    session = Session(config)
+    session = Session.from_config(config)
     session.exitstatus = ExitCode.OK
     initstate = 0
     try:
@@ -202,7 +245,7 @@ def wrap_session(config, doit):
             session.exitstatus = ExitCode.TESTS_FAILED
         except (KeyboardInterrupt, exit.Exception):
             excinfo = _pytest._code.ExceptionInfo.from_current()
-            exitstatus = ExitCode.INTERRUPTED
+            exitstatus = ExitCode.INTERRUPTED  # type: Union[int, ExitCode]
             if isinstance(excinfo.value, exit.Exception):
                 if excinfo.value.returncode is not None:
                     exitstatus = excinfo.value.returncode
@@ -212,7 +255,7 @@ def wrap_session(config, doit):
                     )
             config.hook.pytest_keyboard_interrupt(excinfo=excinfo)
             session.exitstatus = exitstatus
-        except:  # noqa
+        except BaseException:
             session.exitstatus = ExitCode.INTERNAL_ERROR
             excinfo = _pytest._code.ExceptionInfo.from_current()
             try:
@@ -222,25 +265,31 @@ def wrap_session(config, doit):
                     session.exitstatus = exc.returncode
                 sys.stderr.write("{}: {}\n".format(type(exc).__name__, exc))
             else:
-                if excinfo.errisinstance(SystemExit):
+                if isinstance(excinfo.value, SystemExit):
                     sys.stderr.write("mainloop: caught unexpected SystemExit!\n")
 
     finally:
-        excinfo = None  # Explicitly break reference cycle.
+        # Explicitly break reference cycle.
+        excinfo = None  # type: ignore
         session.startdir.chdir()
         if initstate >= 2:
-            config.hook.pytest_sessionfinish(
-                session=session, exitstatus=session.exitstatus
-            )
+            try:
+                config.hook.pytest_sessionfinish(
+                    session=session, exitstatus=session.exitstatus
+                )
+            except exit.Exception as exc:
+                if exc.returncode is not None:
+                    session.exitstatus = exc.returncode
+                sys.stderr.write("{}: {}\n".format(type(exc).__name__, exc))
         config._ensure_unconfigure()
     return session.exitstatus
 
 
-def pytest_cmdline_main(config):
+def pytest_cmdline_main(config: Config) -> Union[int, ExitCode]:
     return wrap_session(config, _main)
 
 
-def _main(config, session):
+def _main(config: Config, session: "Session") -> Optional[Union[int, ExitCode]]:
     """ default command line protocol for initialization, session,
     running tests and reporting. """
     config.hook.pytest_collection(session=session)
@@ -250,13 +299,14 @@ def _main(config, session):
         return ExitCode.TESTS_FAILED
     elif session.testscollected == 0:
         return ExitCode.NO_TESTS_COLLECTED
+    return None
 
 
-def pytest_collection(session):
-    return session.perform_collect()
+def pytest_collection(session: "Session") -> None:
+    session.perform_collect()
 
 
-def pytest_runtestloop(session):
+def pytest_runtestloop(session: "Session") -> bool:
     if session.testsfailed and not session.config.option.continue_on_collection_errors:
         raise session.Interrupted(
             "%d error%s during collection"
@@ -276,7 +326,7 @@ def pytest_runtestloop(session):
     return True
 
 
-def _in_venv(path):
+def _in_venv(path: py.path.local) -> bool:
     """Attempts to detect if ``path`` is the root of a Virtual Environment by
     checking for the existence of the appropriate activate script"""
     bindir = path.join("Scripts" if sys.platform.startswith("win") else "bin")
@@ -293,7 +343,7 @@ def _in_venv(path):
     return any([fname.basename in activates for fname in bindir.listdir()])
 
 
-def pytest_ignore_collect(path, config):
+def pytest_ignore_collect(path: py.path.local, config: Config) -> Optional[bool]:
     ignore_paths = config._getconftest_pathlist("collect_ignore", path=path.dirpath())
     ignore_paths = ignore_paths or []
     excludeopt = config.getoption("ignore")
@@ -317,11 +367,10 @@ def pytest_ignore_collect(path, config):
     allow_in_venv = config.getoption("collect_in_virtualenv")
     if not allow_in_venv and _in_venv(path):
         return True
-
-    return False
+    return None
 
 
-def pytest_collection_modifyitems(items, config):
+def pytest_collection_modifyitems(items: List[nodes.Item], config: Config) -> None:
     deselect_prefixes = tuple(config.getoption("deselect") or [])
     if not deselect_prefixes:
         return
@@ -339,18 +388,6 @@ def pytest_collection_modifyitems(items, config):
         items[:] = remaining
 
 
-class FSHookProxy:
-    def __init__(self, fspath, pm, remove_mods):
-        self.fspath = fspath
-        self.pm = pm
-        self.remove_mods = remove_mods
-
-    def __getattr__(self, name):
-        x = self.pm.subset_hook_caller(name, remove_plugins=self.remove_mods)
-        self.__dict__[name] = x
-        return x
-
-
 class NoMatch(Exception):
     """ raised if matching cannot locate a matching names. """
 
@@ -367,9 +404,9 @@ class Failed(Exception):
 
 @attr.s
 class _bestrelpath_cache(dict):
-    path = attr.ib()
+    path = attr.ib(type=py.path.local)
 
-    def __missing__(self, path: str) -> str:
+    def __missing__(self, path: py.path.local) -> str:
         r = self.path.bestrelpath(path)  # type: str
         self[path] = r
         return r
@@ -382,30 +419,46 @@ class Session(nodes.FSCollector):
     _setupstate = None  # type: SetupState
     # Set on the session by fixtures.pytest_sessionstart.
     _fixturemanager = None  # type: FixtureManager
+    exitstatus = None  # type: Union[int, ExitCode]
 
-    def __init__(self, config):
+    def __init__(self, config: Config) -> None:
         nodes.FSCollector.__init__(
             self, config.rootdir, parent=None, config=config, session=self, nodeid=""
         )
         self.testsfailed = 0
         self.testscollected = 0
-        self.shouldstop = False
-        self.shouldfail = False
+        self.shouldstop = False  # type: Union[bool, str]
+        self.shouldfail = False  # type: Union[bool, str]
         self.trace = config.trace.root.get("collection")
-        self._norecursepatterns = config.getini("norecursedirs")
         self.startdir = config.invocation_dir
-        self._initialpaths = frozenset()
+        self._initialpaths = frozenset()  # type: FrozenSet[py.path.local]
+
         # Keep track of any collected nodes in here, so we don't duplicate fixtures
-        self._node_cache = {}
+        self._collection_node_cache1 = (
+            {}
+        )  # type: Dict[py.path.local, Sequence[nodes.Collector]]
+        self._collection_node_cache2 = (
+            {}
+        )  # type: Dict[Tuple[Type[nodes.Collector], py.path.local], nodes.Collector]
+        self._collection_node_cache3 = (
+            {}
+        )  # type: Dict[Tuple[Type[nodes.Collector], str], CollectReport]
+
+        # Dirnames of pkgs with dunder-init files.
+        self._collection_pkg_roots = {}  # type: Dict[str, Package]
+
         self._bestrelpathcache = _bestrelpath_cache(
             config.rootdir
-        )  # type: Dict[str, str]
-        # Dirnames of pkgs with dunder-init files.
-        self._pkg_roots = {}
+        )  # type: Dict[py.path.local, str]
 
         self.config.pluginmanager.register(self, name="session")
 
-    def __repr__(self):
+    @classmethod
+    def from_config(cls, config: Config) -> "Session":
+        session = cls._create(config)  # type: Session
+        return session
+
+    def __repr__(self) -> str:
         return "<%s %s exitstatus=%r testsfailed=%d testscollected=%d>" % (
             self.__class__.__name__,
             self.name,
@@ -414,19 +467,21 @@ class Session(nodes.FSCollector):
             self.testscollected,
         )
 
-    def _node_location_to_relpath(self, node_path: str) -> str:
+    def _node_location_to_relpath(self, node_path: py.path.local) -> str:
         # bestrelpath is a quite slow function
         return self._bestrelpathcache[node_path]
 
     @hookimpl(tryfirst=True)
-    def pytest_collectstart(self):
+    def pytest_collectstart(self) -> None:
         if self.shouldfail:
             raise self.Failed(self.shouldfail)
         if self.shouldstop:
             raise self.Interrupted(self.shouldstop)
 
     @hookimpl(tryfirst=True)
-    def pytest_runtest_logreport(self, report):
+    def pytest_runtest_logreport(
+        self, report: Union[TestReport, CollectReport]
+    ) -> None:
         if report.failed and not hasattr(report, "wasxfail"):
             self.testsfailed += 1
             maxfail = self.config.getvalue("maxfail")
@@ -435,24 +490,27 @@ class Session(nodes.FSCollector):
 
     pytest_collectreport = pytest_runtest_logreport
 
-    def isinitpath(self, path):
+    def isinitpath(self, path: py.path.local) -> bool:
         return path in self._initialpaths
 
-    def gethookproxy(self, fspath):
-        # check if we have the common case of running
-        # hooks with all conftest.py files
-        pm = self.config.pluginmanager
-        my_conftestmodules = pm._getconftestmodules(fspath)
-        remove_mods = pm._conftest_plugins.difference(my_conftestmodules)
-        if remove_mods:
-            # one or more conftests are not in use at this fspath
-            proxy = FSHookProxy(fspath, pm, remove_mods)
-        else:
-            # all plugins are active for this fspath
-            proxy = self.config.hook
-        return proxy
+    def gethookproxy(self, fspath: py.path.local):
+        return super()._gethookproxy(fspath)
 
-    def perform_collect(self, args=None, genitems=True):
+    @overload
+    def perform_collect(
+        self, args: Optional[Sequence[str]] = ..., genitems: "Literal[True]" = ...
+    ) -> Sequence[nodes.Item]:
+        raise NotImplementedError()
+
+    @overload  # noqa: F811
+    def perform_collect(  # noqa: F811
+        self, args: Optional[Sequence[str]] = ..., genitems: bool = ...
+    ) -> Sequence[Union[nodes.Item, nodes.Collector]]:
+        raise NotImplementedError()
+
+    def perform_collect(  # noqa: F811
+        self, args: Optional[Sequence[str]] = None, genitems: bool = True
+    ) -> Sequence[Union[nodes.Item, nodes.Collector]]:
         hook = self.config.hook
         try:
             items = self._perform_collect(args, genitems)
@@ -465,19 +523,33 @@ class Session(nodes.FSCollector):
         self.testscollected = len(items)
         return items
 
-    def _perform_collect(self, args, genitems):
+    @overload
+    def _perform_collect(
+        self, args: Optional[Sequence[str]], genitems: "Literal[True]"
+    ) -> List[nodes.Item]:
+        raise NotImplementedError()
+
+    @overload  # noqa: F811
+    def _perform_collect(  # noqa: F811
+        self, args: Optional[Sequence[str]], genitems: bool
+    ) -> Union[List[Union[nodes.Item]], List[Union[nodes.Item, nodes.Collector]]]:
+        raise NotImplementedError()
+
+    def _perform_collect(  # noqa: F811
+        self, args: Optional[Sequence[str]], genitems: bool
+    ) -> Union[List[Union[nodes.Item]], List[Union[nodes.Item, nodes.Collector]]]:
         if args is None:
             args = self.config.args
         self.trace("perform_collect", self, args)
         self.trace.root.indent += 1
-        self._notfound = []
-        initialpaths = []
-        self._initialparts = []
-        self.items = items = []
+        self._notfound = []  # type: List[Tuple[str, NoMatch]]
+        initialpaths = []  # type: List[py.path.local]
+        self._initial_parts = []  # type: List[Tuple[py.path.local, List[str]]]
+        self.items = items = []  # type: List[nodes.Item]
         for arg in args:
-            parts = self._parsearg(arg)
-            self._initialparts.append(parts)
-            initialpaths.append(parts[0])
+            fspath, parts = self._parsearg(arg)
+            self._initial_parts.append((fspath, parts))
+            initialpaths.append(fspath)
         self._initialpaths = frozenset(initialpaths)
         rep = collect_one_node(self)
         self.ihook.pytest_collectreport(report=rep)
@@ -496,26 +568,29 @@ class Session(nodes.FSCollector):
                     self.items.extend(self.genitems(node))
             return items
 
-    def collect(self):
-        for initialpart in self._initialparts:
-            self.trace("processing argument", initialpart)
+    def collect(self) -> Iterator[Union[nodes.Item, nodes.Collector]]:
+        for fspath, parts in self._initial_parts:
+            self.trace("processing argument", (fspath, parts))
             self.trace.root.indent += 1
             try:
-                yield from self._collect(initialpart)
-            except NoMatch:
-                report_arg = "::".join(map(str, initialpart))
+                yield from self._collect(fspath, parts)
+            except NoMatch as exc:
+                report_arg = "::".join((str(fspath), *parts))
                 # we are inside a make_report hook so
                 # we cannot directly pass through the exception
-                self._notfound.append((report_arg, sys.exc_info()[1]))
+                self._notfound.append((report_arg, exc))
 
             self.trace.root.indent -= 1
+        self._collection_node_cache1.clear()
+        self._collection_node_cache2.clear()
+        self._collection_node_cache3.clear()
+        self._collection_pkg_roots.clear()
 
-    def _collect(self, arg):
+    def _collect(
+        self, argpath: py.path.local, names: List[str]
+    ) -> Iterator[Union[nodes.Item, nodes.Collector]]:
         from _pytest.python import Package
 
-        names = arg[:]
-        argpath = names.pop(0)
-
         # Start with a Session root, and delve to argpath item (dir or file)
         # and stack all Packages found on the way.
         # No point in finding packages when collecting doctests
@@ -528,20 +603,20 @@ class Session(nodes.FSCollector):
                 if parent.isdir():
                     pkginit = parent.join("__init__.py")
                     if pkginit.isfile():
-                        if pkginit not in self._node_cache:
+                        if pkginit not in self._collection_node_cache1:
                             col = self._collectfile(pkginit, handle_dupes=False)
                             if col:
                                 if isinstance(col[0], Package):
-                                    self._pkg_roots[parent] = col[0]
+                                    self._collection_pkg_roots[str(parent)] = col[0]
                                 # always store a list in the cache, matchnodes expects it
-                                self._node_cache[col[0].fspath] = [col[0]]
+                                self._collection_node_cache1[col[0].fspath] = [col[0]]
 
         # If it's a directory argument, recurse and look for any Subpackages.
         # Let the Package collector deal with subnodes, don't collect here.
         if argpath.check(dir=1):
-            assert not names, "invalid arg {!r}".format(arg)
+            assert not names, "invalid arg {!r}".format((argpath, names))
 
-            seen_dirs = set()
+            seen_dirs = set()  # type: Set[py.path.local]
             for path in argpath.visit(
                 fil=self._visit_filter, rec=self._recurse, bf=True, sort=True
             ):
@@ -554,36 +629,37 @@ class Session(nodes.FSCollector):
                         for x in self._collectfile(pkginit):
                             yield x
                             if isinstance(x, Package):
-                                self._pkg_roots[dirpath] = x
-                if dirpath in self._pkg_roots:
+                                self._collection_pkg_roots[str(dirpath)] = x
+                if str(dirpath) in self._collection_pkg_roots:
                     # Do not collect packages here.
                     continue
 
                 for x in self._collectfile(path):
                     key = (type(x), x.fspath)
-                    if key in self._node_cache:
-                        yield self._node_cache[key]
+                    if key in self._collection_node_cache2:
+                        yield self._collection_node_cache2[key]
                     else:
-                        self._node_cache[key] = x
+                        self._collection_node_cache2[key] = x
                         yield x
         else:
             assert argpath.check(file=1)
 
-            if argpath in self._node_cache:
-                col = self._node_cache[argpath]
+            if argpath in self._collection_node_cache1:
+                col = self._collection_node_cache1[argpath]
             else:
-                collect_root = self._pkg_roots.get(argpath.dirname, self)
+                collect_root = self._collection_pkg_roots.get(argpath.dirname, self)
                 col = collect_root._collectfile(argpath, handle_dupes=False)
                 if col:
-                    self._node_cache[argpath] = col
+                    self._collection_node_cache1[argpath] = col
             m = self.matchnodes(col, names)
             # If __init__.py was the only file requested, then the matched node will be
             # the corresponding Package, and the first yielded item will be the __init__
             # Module itself, so just use that. If this special case isn't taken, then all
             # the files in the package will be yielded.
             if argpath.basename == "__init__.py":
+                assert isinstance(m[0], nodes.Collector)
                 try:
-                    yield next(m[0].collect())
+                    yield next(iter(m[0].collect()))
                 except StopIteration:
                     # The package collects nothing with only an __init__.py
                     # file in it, which gets ignored by the default
@@ -592,46 +668,12 @@ class Session(nodes.FSCollector):
                 return
             yield from m
 
-    def _collectfile(self, path, handle_dupes=True):
-        assert (
-            path.isfile()
-        ), "{!r} is not a file (isdir={!r}, exists={!r}, islink={!r})".format(
-            path, path.isdir(), path.exists(), path.islink()
-        )
-        ihook = self.gethookproxy(path)
-        if not self.isinitpath(path):
-            if ihook.pytest_ignore_collect(path=path, config=self.config):
-                return ()
-
-        if handle_dupes:
-            keepduplicates = self.config.getoption("keepduplicates")
-            if not keepduplicates:
-                duplicate_paths = self.config.pluginmanager._duplicatepaths
-                if path in duplicate_paths:
-                    return ()
-                else:
-                    duplicate_paths.add(path)
-
-        return ihook.pytest_collect_file(path=path, parent=self)
-
-    def _recurse(self, dirpath):
-        if dirpath.basename == "__pycache__":
-            return False
-        ihook = self.gethookproxy(dirpath.dirpath())
-        if ihook.pytest_ignore_collect(path=dirpath, config=self.config):
-            return False
-        for pat in self._norecursepatterns:
-            if dirpath.check(fnmatch=pat):
-                return False
-        ihook = self.gethookproxy(dirpath)
-        ihook.pytest_collect_directory(path=dirpath, parent=self)
-        return True
-
     @staticmethod
-    def _visit_filter(f):
-        return f.check(file=1)
+    def _visit_filter(f: py.path.local) -> bool:
+        # TODO: Remove type: ignore once `py` is typed.
+        return f.check(file=1)  # type: ignore
 
-    def _tryconvertpyarg(self, x):
+    def _tryconvertpyarg(self, x: str) -> str:
         """Convert a dotted module name to path."""
         try:
             spec = importlib.util.find_spec(x)
@@ -640,30 +682,31 @@ class Session(nodes.FSCollector):
         # ValueError: not a module name
         except (AttributeError, ImportError, ValueError):
             return x
-        if spec is None or spec.origin in {None, "namespace"}:
+        if spec is None or spec.origin is None or spec.origin == "namespace":
             return x
         elif spec.submodule_search_locations:
             return os.path.dirname(spec.origin)
         else:
             return spec.origin
 
-    def _parsearg(self, arg):
+    def _parsearg(self, arg: str) -> Tuple[py.path.local, List[str]]:
         """ return (fspath, names) tuple after checking the file exists. """
-        parts = str(arg).split("::")
+        strpath, *parts = str(arg).split("::")
         if self.config.option.pyargs:
-            parts[0] = self._tryconvertpyarg(parts[0])
-        relpath = parts[0].replace("/", os.sep)
-        path = self.config.invocation_dir.join(relpath, abs=True)
-        if not path.check():
+            strpath = self._tryconvertpyarg(strpath)
+        relpath = strpath.replace("/", os.sep)
+        fspath = self.config.invocation_dir.join(relpath, abs=True)
+        if not fspath.check():
             if self.config.option.pyargs:
                 raise UsageError(
                     "file or package not found: " + arg + " (missing __init__.py?)"
                 )
             raise UsageError("file not found: " + arg)
-        parts[0] = path.realpath()
-        return parts
+        return (fspath, parts)
 
-    def matchnodes(self, matching, names):
+    def matchnodes(
+        self, matching: Sequence[Union[nodes.Item, nodes.Collector]], names: List[str],
+    ) -> Sequence[Union[nodes.Item, nodes.Collector]]:
         self.trace("matchnodes", matching, names)
         self.trace.root.indent += 1
         nodes = self._matchnodes(matching, names)
@@ -674,13 +717,15 @@ class Session(nodes.FSCollector):
             raise NoMatch(matching, names[:1])
         return nodes
 
-    def _matchnodes(self, matching, names):
+    def _matchnodes(
+        self, matching: Sequence[Union[nodes.Item, nodes.Collector]], names: List[str],
+    ) -> Sequence[Union[nodes.Item, nodes.Collector]]:
         if not matching or not names:
             return matching
         name = names[0]
         assert name
         nextnames = names[1:]
-        resultnodes = []
+        resultnodes = []  # type: List[Union[nodes.Item, nodes.Collector]]
         for node in matching:
             if isinstance(node, nodes.Item):
                 if not names:
@@ -688,11 +733,11 @@ class Session(nodes.FSCollector):
                 continue
             assert isinstance(node, nodes.Collector)
             key = (type(node), node.nodeid)
-            if key in self._node_cache:
-                rep = self._node_cache[key]
+            if key in self._collection_node_cache3:
+                rep = self._collection_node_cache3[key]
             else:
                 rep = collect_one_node(node)
-                self._node_cache[key] = rep
+                self._collection_node_cache3[key] = rep
             if rep.passed:
                 has_matched = False
                 for x in rep.result:
@@ -711,7 +756,9 @@ class Session(nodes.FSCollector):
                 node.ihook.pytest_collectreport(report=rep)
         return resultnodes
 
-    def genitems(self, node):
+    def genitems(
+        self, node: Union[nodes.Item, nodes.Collector]
+    ) -> Iterator[nodes.Item]:
         self.trace("genitems", node)
         if isinstance(node, nodes.Item):
             node.ihook.pytest_itemcollected(item=node)
